/*
 * Copyright 2012 Google Inc.
 *
 * Use of this source code is governed by a BSD-style license that can be
 * found in the LICENSE file.
 */

#include "SkThreadUtils.h"

bool SkThread::setProcessorAffinity(unsigned int processor) {
    return false;
}
