/*
 * Copyright 2006 The Android Open Source Project
 *
 * Use of this source code is governed by a BSD-style license that can be
 * found in the LICENSE file.
 */

#include "SkXMLParser.h"

bool SkXMLParser::parse(SkStream& docStream)
{
    return false;
}

bool SkXMLParser::parse(const char doc[], size_t len)
{
    return false;
}

void SkXMLParser::GetNativeErrorString(int error, SkString* str)
{

}
