/*
 *  Copyright (c) 2015 The WebRTC project authors. All Rights Reserved.
 *
 *  Use of this source code is governed by a BSD-style license
 *  that can be found in the LICENSE file in the root of the source
 *  tree. An additional intellectual property rights grant can be found
 *  in the file PATENTS.  All contributing project authors may
 *  be found in the AUTHORS file in the root of the source tree.
 */

#include "webrtc/modules/audio_processing/beamformer/mock_nonlinear_beamformer.h"

#include <vector>

namespace webrtc {

MockNonlinearBeamformer::MockNonlinearBeamformer(
    const std::vector<Point>& array_geometry)
    : NonlinearBeamformer(array_geometry) {
}

}  // namespace webrtc
